<?php

namespace App\Http\Controllers\Api;

use App\{Models\AgentProfile, Models\PauseReason};
use App\Http\Controllers\Controller;
use App\Http\Requests\AgentProfiles\{StoreAgentProfile, UpdateAgentProfile};
use App\Utils\TenantUtils;
use Illuminate\Http\Request;

class AgentProfileController extends Controller
{
    public function index(Request $request)
    {
        $this->authorize('viewAny', AgentProfile::class);

        return [
            'data' => AgentProfile::with('pauses')
                ->byTenant(TenantUtils::getTenantId($request))
                ->global()->default()->get()
        ];
    }

    public function store(StoreAgentProfile $request)
    {
        $validatedData = $request->validated();

        $this->authorize('create', AgentProfile::class);

        $agentProfile = AgentProfile::create($validatedData);

        $pauses = collect($validatedData['pauses']);

        $agentProfile->pauses()->createMany($pauses);

        $agentProfile->load('pauses');

        return ['data' => $agentProfile];
    }

    public function show(Request $request, $id)
    {
        $agentProfile = AgentProfile::with('pauses')->findOrFail($id);

        $this->authorize('viewUpdatedestroy', $agentProfile);

        return ['data' => $agentProfile];
    }

    public function update(UpdateAgentProfile $request, $id)
    {
        $defaultPauses = PauseReason::DEFAULTPAUSES;

        $validatedData = $request->validated();

        $agentProfile = AgentProfile::with('pauses')->findOrFail($id);

        $this->authorize('viewUpdatedestroy', $agentProfile);

        //default profile cannot be updated
        if ($agentProfile->id !== 1) {
            $agentProfile->update($validatedData);
        }

        $pauses = collect($validatedData['pauses']);

        $pauseIds = $agentProfile->pauses()
            ->when($agentProfile->id === 1, function ($query) {
                return $query->editable();
            })
            ->pluck('id')->toArray();

        $currentPauses = $pauses->whereIn('id', $pauseIds);

        $agentProfile->pauses()->whereNotIn('id', $currentPauses->pluck('id')->toArray())
            ->when($agentProfile->id === 1, function ($query) use ($defaultPauses) {
                return $query->whereNotIn('name', $defaultPauses);
            })
        ->delete();

        $currentPauses->each(function ($item) use ($pauses, $agentProfile) {
            $pause = $pauses->where('id', $item['id'])->first();

            if ($pause) {
                $agentProfile->pauses()->updateOrCreate(
                    ['id' => $item['id']],
                    ['name' => $item['name']]
                );
            }
        });

        //don't save default pauses on default profile
        $newPauses = $pauses->whereNotIn('id', $pauseIds)->when($agentProfile->id === 1, function ($query) use ($defaultPauses) {
           return $query->whereNotIn('name', $defaultPauses);
        });

        $agentProfile->pauses()->createMany($newPauses);

        $agentProfile->load('pauses');

        return ['data' => $agentProfile];
    }

    public function destroy(Request $request, $id)
    {
        $agentProfile = AgentProfile::with('agents')->findOrFail($id);

        $this->authorize('viewUpdatedestroy', $agentProfile);

        if ($agentProfile->agents()->count() > 0) {
            return response()->json(['message' => __('pause_profile_cannot_delete')], 422);
        }

        $agentProfile->delete();

        return response()->json([], 204);
    }
}
