<?php

namespace App\Http\Controllers\Api;

use App\{Models\BroadcastCampaign, Models\BroadcastCampaignContact, Models\ContactList, Utils\VitalPBXApiUtils};
use App\Http\Controllers\Controller;
use App\Http\Requests\BroadcastCampaigns\StoreBroadcastCampaign;
use App\Utils\TenantUtils;
use Illuminate\Http\Request;
use Illuminate\Validation\Rule;

class BroadcastCampaignController extends Controller
{
    public function index(Request $request)
    {
        $this->authorize('viewAny', BroadcastCampaign::class);

        return ['data' => BroadcastCampaign::byTenant(TenantUtils::getTenantId($request))
            ->orderBy('updated_at', 'desc')
            ->get()
        ];
    }

    public function store(StoreBroadcastCampaign $request)
    {
        $this->authorize('create', BroadcastCampaign::class);

        $validatedData = $request->validated();

        $broadcastCampaign = new BroadcastCampaign($validatedData);

        if ($request->has('active_on_create') && $validatedData['active_on_create']) {
            $broadcastCampaign->status = 'active';
        }

        $broadcastCampaign->save();

        $broadcastCampaign->contacts()->attach(ContactList::find($validatedData['contact_list_id'])->contacts);

        $broadcastCampaign->refresh();
        return ['data' => $broadcastCampaign];
    }

    public function show(BroadcastCampaign $broadcastCampaign)
    {
        $this->authorize('broadcastCampaign', $broadcastCampaign);

        return ['data' => $broadcastCampaign];
    }

    public function update(StoreBroadcastCampaign $request, BroadcastCampaign $broadcastCampaign)
    {
        $this->authorize('broadcastCampaign', $broadcastCampaign);

        $validatedData = $request->validated();

        if ($broadcastCampaign->status == 'active') {
            return response()->json(['message' => __('campaign_cannot_update')], 422);
        }

        $broadcastCampaign->update($validatedData);

        $broadcastCampaign->contacts()->sync(ContactList::find($validatedData['contact_list_id'])->contacts);

        return ['data' => $broadcastCampaign];
    }

    public function destroy(BroadcastCampaign $broadcastCampaign)
    {
       $this->authorize('broadcastCampaign', $broadcastCampaign);

       if ($broadcastCampaign->status == 'active') {
           return response()->json(['message' => __('campaign_cannot_delete')], 422);
       }

       $broadcastCampaign->delete();

       return response()->json([], 204);
    }

    public function status(Request $request, BroadcastCampaign $broadcastCampaign)
    {
        $this->authorize('broadcastCampaign', $broadcastCampaign);

        $validatedData = $request->validate([
            'status' => ['required', Rule::in(['active', 'inactive', 'finished'])],
        ]);

        $broadcastCampaign->update($validatedData);

        BroadcastCampaignContact::where('broadcast_campaign_id', $broadcastCampaign->id)
            ->where('status', '=', 'queued')
            ->update(['status' => 'pending', 'called_at' => null]);

        return ['data' => $broadcastCampaign];
    }

    public function testCall(Request $request) {
        // TODO: check if this line is necessary
        // $this->authorize('broadcastCampaign', BroadcastCampaign::class);

        $validatedData = $request->validate([
            'name' => ['required'],
            'cos_id' => ['required'],
            'destination_category_id' => ['required'],
            'destination_id' => ['required'],
            'cid_name' => ['nullable'],
            'cid_number' => ['nullable'],
            'number' => ['required'],
            'tenant_id' => ['nullable'],
        ]);

        $cid_name = 'S. Dialer - Test Call';
        if ($validatedData['name']) {
            $cid_name = $validatedData['name'];
        } if ($validatedData['cid_name']) {
            $cid_name = $validatedData['cid_name'];
        }

        // define how long to wait for call to be answered (in secs.)
        $timeout = 30;
        $params = [
            'number' => $validatedData['number'],
            'cos_id' => $validatedData['cos_id'],
            'destination_category_id' => $validatedData['destination_category_id'],
            'destination_id' => $validatedData['destination_id'],
            'cid_name' => $cid_name,
            'cid_number' => $validatedData['cid_number'],
            'timeout' => $timeout,
        ];
        $httpRequestTimeout = $timeout + 10;

        $tenantId = $validatedData['tenant_id'];
        if (auth()->user() && is_null($tenantId)) {
            $tenantId = auth()->user()->tenant_id;
        }

        $response = VitalPBXApiUtils::request(
            'core/dialer_call',
            $tenantId,
            method: 'POST',
            params: $params,
            timeout: $httpRequestTimeout
        );

        if (is_array($response) && isset($response['data'])) {
            return response()->json([
                'message' => __('call_test_success'),
            ]);
        }
        return response()->json([
            'message' => __('call_test_failed'),
        ], 422);

    }

    public function recycle(BroadcastCampaign $broadcastCampaign)
    {
        $this->authorize('broadcastCampaign', $broadcastCampaign);

        if ($broadcastCampaign->status != 'finished') {
            return response()->json(['message' => __('campaign_cannot_recycle')], 422);
        }

        BroadcastCampaignContact::where('broadcast_campaign_id', $broadcastCampaign->id)
            ->update(['status' => 'pending', 'called_at' => null]);

        $broadcastCampaign->update(['status' => 'inactive']);
        return ['message' => __('campaign_recycle_success')];

    }
}
