<?php

namespace App\Http\Controllers\Api;

use App\Http\Controllers\Controller;
use App\Models\TimezoneSetting;
use App\Utils\{VitalPBXApiUtils};
use Carbon\Carbon;
use Illuminate\Http\Request;

class CallHistoryController extends Controller
{
    public function index(Request $request)
    {
        try {
            $user = auth()->user();
            $extension = $user->extension;

            if (is_null($extension)) {
                throw new UnexpectedValueException(
                    __('you_do_not_have_an_extension_assigned'),
                    422
                );
            }

            $timezone = TimezoneSetting::getSettings($user->tenant_id)->timezone;

            $historyLimit = 50;

            $startTimestamp = $request->input('start_date', Carbon::now()->subMonths(2)->timestamp);
            $endTimestamp = $request->input('end_date', Carbon::now()->timestamp);

            $query = [
                'extensions' => $extension,
                'limit' => $historyLimit,
                'start_date' => $startTimestamp,
                'end_date' => $endTimestamp
            ];
            $response = VitalPBXApiUtils::request(
                'cdr',
                $user->tenant_id,
                'GET',
                null,
                null,
                30,
                $query
            );

            $calls = [];
            $lostCalls = 0;

            if (isset($response['data']['result'])) {
                $data = $this->parseCallsData($response['data']['result'], $timezone);

                $calls = collect($data->calls)
                    ->unique('uniqueid')
                    ->sortByDesc('calldate')
                    ->values();
                $lostCalls = $data->lost_calls;
            }

            $responseData = [
                'calls' => $calls,
                'lost_calls' => $lostCalls
            ];
            return response()->json($responseData);
        } catch (Exception | UnexpectedValueException $e) {
            $errorResponse = [
                'error' => true,
                'code' => $e->getCode(),
                'message' => $e->getMessage()
            ];

            return response()->json($errorResponse, $e->getCode());
        }
    }

    public function formatCallerID ($str = '') {
        $name = '';
        $re = '/([A-Z])\w+/m';

        if (is_string($str) && !empty($str)) {
            preg_match_all($re, $str, $matches, PREG_SET_ORDER, 0);
            foreach ($matches as $key => $value) {
                $name.= $value[0];
                if (array_key_last($matches) !== $key) {
                    $name.= ' ';
                }
            }
        }
        return $name;
    }

    public function parseCallsData ($data, $timezone = 'UTC'): object
    {
        $user = auth()->user();
        $calls = [];
        $lostCalls = 0;
        foreach ($data as $key => $value) {
            $number = $value['source'] !== $user->extension ? $value['source'] : $value['destination'];
            $call_info = (object) [
                'name' => $value['source'] !== $user->extension ? $this->formatCallerID($value['clid']) : $value['destination'],
                'original_name' =>  $value['source'] !== $user->extension ? $this->formatCallerID($value['clid']) : $value['destination'],
                'number' => $number ?? '',
                'calldate' => Carbon::createFromFormat('Y-m-d H:i:s', $value['calldate'], 'UTC')->setTimezone($timezone)->format('Y-m-d H:i:s'),
                'originator' => $value['destination'] === $user->extension ? 'remote' : 'local',
                'duration' => $value['duration'],
                'disposition' => $value['disposition'],
                'uniqueid' => $value['uniqueid'],
                'calltype' => $value['calltype'],
                'recfile' => $value['recfile'],
                'tenant' => $value['tenant'],
                'trunk' => $value['trunk'],
            ];

            $calls[] = $call_info;
        }
        return (object) [
            'calls' => $calls,
            'lost_calls' => $lostCalls
        ];
    }

}
