<?php

namespace App\Http\Controllers\Api;

use App\{Models\Campaign, Models\CampaignContact, Models\CampaignResult, Models\ContactList};
use App\Http\Controllers\Controller;
use App\Http\Requests\Campaigns\{StatusUpdateRequest, StoreCampaign, UpdateCampaign};
use App\Http\Resources\Campaign as CampaignResource;
use App\Utils\TenantUtils;
use Illuminate\Http\Request;
use Illuminate\Validation\Rule;

class CampaignController extends Controller
{
    public function index(Request $request)
    {
        $this->authorize('viewAny', Campaign::class);

        $user = auth()->user()->load('systemPermission');

        if ($user->isSuperAdmin() || (is_null($user->systemPermission) || $user->systemPermission->campaigns()->count() === 0)) {
            $campaigns = Campaign::with('dispositionGroup')
                ->byTenant(TenantUtils::getTenantId($request))
                ->orderBy('updated_at', 'desc')
                ->ofStatus($request->input('status'))
                ->get();
        } else {
            $campaigns = $user->systemPermission->campaigns ?? [];
        }

        return ['data' => $campaigns];
    }

    public function store(StoreCampaign $request)
    {
        $this->authorize('create', Campaign::class);

        $validatedData = $request->validated();

        $campaign = new Campaign($validatedData);

        if ($validatedData['active_on_create']) {
            $campaign->status = 'active';
        }

        $campaign->save();
        
        $campaign->contacts()->attach(ContactList::find($validatedData['contact_list_id'])->contacts);

        $campaign->refresh();

        return new CampaignResource($campaign);
    }

    public function update(UpdateCampaign $request, Campaign $campaign)
    {
        $this->authorize('campaign', $campaign);

        $validatedData = $request->validated();

        $campaign->update($validatedData);
        
        $campaign->contacts()->sync(ContactList::find($validatedData['contact_list_id'])->contacts);

        return new CampaignResource($campaign);
    }

    public function show(Request $request, $id)
    {
        $campaign = Campaign::with('dispositionGroup.dispositions')
            ->withCount(['contacts as total_contacts', 'contacts as finished_contacts' => function ($query) {
                $query->where('status', 'finished');
            }])
            ->findOrFail($id);

        $this->authorize('campaign', $campaign);

        $campaign->call_later_contacts = CampaignResult::byCampaign($campaign->id)
            ->whereHas('disposition', function ($query) {
                $query->where('category', 'call_later');
            })->count();

        return new CampaignResource($campaign);
    }

    public function destroy(Request $request, Campaign $campaign)
    {
        $this->authorize('campaign', $campaign);
    
        $response = $this->checkActiveCampaign($campaign->status, 'delete');

        if ($response) return $response;

        $campaign->delete();

        return response()->json([], 204);
    }

    public function status(StatusUpdateRequest $request, Campaign $campaign)
    {
        $validatedData = $request->validated();

        $this->authorize('campaign', $campaign);

        // improve: message
        if ($campaign->status == 'finished' && $validatedData['status'] === 'active') {
            return response()->json(['message' => __('campaign_cannot_update_status')], 422);
        }
        
        $campaign->update(['status' => $validatedData['status']]);

        return ['data' => $campaign];
    }

    public function recycle(Request $request, Campaign $campaign)
    {
        $this->authorize('campaign', $campaign);

        $validatedData = $request->validate([
            'category' => ['required', Rule::in(['all', 'call_later'])]
        ]);

        if ($campaign->status != 'finished') {
            return response()->json(['message' => __('campaign_cannot_recycle')], 422);
        }

        if ($validatedData['category'] === 'all') {
            CampaignContact::byCampaign($campaign->id)->update([
                'agent_id' => null, 'call_at' => null, 'status' => 'pending'
            ]);
        }

        if ($validatedData['category'] === 'call_later') {
            //find contacts with call later disposition
            $contactIds = CampaignResult::with('disposition')
                ->byCampaign($campaign->id)
                ->whereHas('disposition', function ($query) {
                    $query->where('category', 'call_later');
                })
                ->pluck('contact_id')
                ->all();

            //remove any scheduled calls
            CampaignContact::byCampaign($campaign->id)
                ->scheduled()
                ->update(['call_at' => null]);

            //recycle only contacts with call later disposition
            CampaignContact::byCampaign($campaign->id)
                ->whereIn('contact_id', $contactIds)
                ->update(['agent_id' => null, 'status' => 'pending']);
        }

        $campaign->update(['status' => 'inactive']);

        return ['message' => __('campaign_recycle_success')];
    }

    private function checkActiveCampaign($status, $action)
    {
        $message = [
            'update' => __('campaign_cannot_update'),
            'delete' => __('campaign_cannot_delete'),
        ];
    
        if ($status === 'active') {
            return response()->json(['message' => $message[$action]], 422);
        } 
    }
}
