<?php

namespace App\Http\Controllers\Api;

use App\Services\AgentConsoleService;
use App\Utils\LicenseUtils;
use BeyondCode\LaravelWebSockets\WebSockets\Exceptions\UnknownAppKey;
use BeyondCode\LaravelWebSockets\Apps\App;
use Illuminate\Support\Facades\Log;
use Ratchet\ConnectionInterface;
use Ratchet\RFC6455\Messaging\MessageInterface;
use Ratchet\WebSocket\MessageComponentInterface;
use JWTAuth;
use Tymon\JWTAuth\Exceptions\{JWTException, TokenExpiredException, TokenInvalidException};

class ConsoleWebSocketHandler implements MessageComponentInterface
{
    const EVENTS = [
        'ping' => 'ping', 'agent_campaigns' => 'agentCampaigns', 'call' => 'callContact',
        'contact_data' => 'sendContactData', 'save_form' => 'saveCampaignResult',
        'login' => 'login', 'logout' => 'logout', 'pause' => 'pause', 'unpause' => 'unpause',
        'dynamic-queues' => 'dynamicQueues', 'queues-status' => 'queuesStatus',
        'index_scheduled_calls' => 'indexScheduledCalls', 'save_scheduled_calls' => 'saveScheduledCalls'
    ];

    public function onOpen(ConnectionInterface $connection)
    {
        $this->verifyAppKey($connection)->generateSocketId($connection);
        //agent won't be able to log in the agent console
        //if a customer tries to manually create more agents
        //than allowed by the license
        if (! LicenseUtils::checkLicenseRestriction(true)) {
            $connection->close();
        }
    }

    public function onClose(ConnectionInterface $connection)
    {
        // TODO: Implement onClose() method.
    }

    public function onError(ConnectionInterface $connection, \Exception $e)
    {
        Log::info($e->getMessage());
        $connection->close();
    }

    public function onMessage(ConnectionInterface $connection, MessageInterface $msg)
    {
        $msg = json_decode($msg);

        //verify valid user token
        try {
            JWTAuth::setToken($msg->token ?? null);
            JWTAuth::getPayload();
            $user = JWTAuth::toUser($msg->token);
            $msg->payload->agent_id = $user->id ?? null;
            $msg->payload->device_id = $user->device_id ?? null;
            $msg->payload->tenant_id = $user->tenant_id ?? null;
            $msg->payload->extension = $user->extension ?? null;
            \Illuminate\Support\Facades\App::setLocale($user->locale);
        } catch (TokenExpiredException | TokenInvalidException | JWTException  | \Exception $e) {
            $user = null;
        }

        if ($user) {
            $func = self::EVENTS[$msg->event];

            $connection->send(json_encode(AgentConsoleService::$func($msg->payload)));
        } else {
            $connection->send(json_encode(['message' => __('console_login_error'), 'event' => $msg->event ?? '', 'error' => true]));
        }
    }

    protected function verifyAppKey(ConnectionInterface $connection)
    {
        $appKey = 'dialer';

        if (! $app = App::findByKey($appKey)) {
            throw new Exceptions\UnknownAppKey($appKey);
        }

        $connection->app = $app;

        return $this;
    }

    protected function generateSocketId(ConnectionInterface $connection)
    {
        $socketId = sprintf('%d.%d', random_int(1, 1000000000), random_int(1, 1000000000));

        $connection->socketId = $socketId;

        return $this;
    }
}
