<?php

namespace App\Http\Controllers\Api;

use App\Exports\ContactsExport;
use App\Http\Controllers\Controller;
use App\Http\Requests\ContactLists\StoreContactList;
use App\Http\Requests\Contacts\StoreContact;
use App\Http\Resources\ContactList as ContactListResource;
use App\Models\ContactList;
use App\Utils\TenantUtils;
use Illuminate\Http\Request;

class ContactListController extends Controller
{
    public function index(Request $request)
    {
        $this->authorize('viewAny', ContactList::class);
        
        return ContactListResource::collection(ContactList::with(['campaigns', 'broadcastCampaigns'])->withCount('contacts')
            ->byTenant(TenantUtils::getTenantId($request))
            ->orderBy('updated_at', 'desc')
            ->get());
    }

    public function store(StoreContactList $request)
    {
        $this->authorize('contactList', ContactList::class);
        
        return new ContactListResource(ContactList::create($request->validated()));
    }

    public function show(Request $request, ContactList $contactList)
    {
        $this->authorize('viewUpdateDestroy', $contactList);
        
        return new ContactListResource($contactList);
    }

    public function update(StoreContactList $request, $id)
    {
        $contactList = ContactList::with('campaigns')->findOrFail($id);

        $this->authorize('viewUpdateDestroy', $contactList);
    
        $response = $this->checkCampaigns($contactList, 'update');

        if ($response) return $response;

        $contactList->update($request->validated());

        return new ContactListResource($contactList);
    }

    public function destroy(Request $request, $id)
    {
        $contactList = ContactList::with('campaigns')->findOrFail($id);

        $this->authorize('viewUpdateDestroy', $contactList);
    
        $response = $this->checkCampaigns($contactList, 'delete');

        if ($response) return $response;

        $contactList->delete();

        return response()->json([], 204);
    }

    public function contacts(Request $request, $id)
    {
        $this->authorize('contactList', ContactList::class);
        
        return ContactList::with('contacts')
            ->byTenant(TenantUtils::getTenantId($request))
            ->findOrFail($id)
            ->contacts()
            ->paginate();
    }

    public function import(StoreContact $request, $id)
    {
        $this->authorize('contactList', ContactList::class);
        
        $validatedData = $request->validated();
        
        $contactList = ContactList::with('campaigns')
            ->byTenant(TenantUtils::getTenantId($request))
            ->findOrFail($id);

        $response = $this->checkCampaigns($contactList, 'import');

        if ($response) return $response;
        
        $functions = [
            'file' => 'App\Imports\FileImport',
            'odoo' => 'App\Imports\OdooImport',
            'salesforce' => 'App\Imports\SalesforceImport'
        ];
        
        $response = $functions[$validatedData['import_type']]::execute($contactList, $validatedData);
        
        if ($response) return $response;

        $contactList->loadCount('contacts');
        
        return new ContactListResource($contactList);
    }

    public function export(Request $request, $id)
    {
        $this->authorize('contactList', ContactList::class);
        
        $contactList = ContactList::byTenant(TenantUtils::getTenantId($request))->findOrFail($id);
    
        return (new ContactsExport($contactList->id))->download('contacts.csv');
    }

    /**
     * validates records with associated campaigns cannot be deleted
     */
    private function checkCampaigns($contactList, $action)
    {
        $message = [
            'import' => __('contact_list_cannot_import'),
            'delete' => __('contact_list_cannot_delete'),
            'update' => __('contact_list_cannot_update'),
        ];
    
        if ($contactList->campaigns()->active()->count() > 0
            || $contactList->broadcastCampaigns()->active()->count() > 0) {
            return response()->json(['message' => $message[$action]], 422);
        } 
    }
}
