<?php

namespace App\Http\Controllers\Api;

use App\Http\Controllers\Controller;
use App\Http\Requests\DispositionGroups\StoreDispositionGroup;
use App\Models\DispositionGroup;
use App\Utils\TenantUtils;
use Illuminate\Http\Request;

class DispositionGroupController extends Controller
{
    public function index(Request $request)
    {
        $this->authorize('viewAny', DispositionGroup::class);
        
        return [
            'data' => DispositionGroup::byTenant(TenantUtils::getTenantId($request))
                ->with('dispositions')
                ->orderBy('updated_at', 'desc')
                ->get()
        ];
    }

    public function store(StoreDispositionGroup $request)
    {
        $this->authorize('create', DispositionGroup::class);
        
        $validatedData = $request->validated();
    
        $dispositionGroup = DispositionGroup::create($validatedData);

        $dispositions = collect($validatedData['dispositions']);

        $dispositions->transform(function ($item) use ($validatedData) {
            $item['tenant_id'] = $validatedData['tenant_id'];
            return $item;
        });
        
        $dispositionGroup->dispositions()->createMany($dispositions);
        
        $dispositionGroup->load('dispositions');
    
        return ['data' => $dispositionGroup];
    }

    public function update(StoreDispositionGroup $request, $id)
    {
        $validatedData = $request->validated();
    
        $dispositionGroup = DispositionGroup::with('campaigns', 'dispositions')->findOrFail($id);

        $this->authorize('updateDestroy', $dispositionGroup);
            
        $response = $this->checkCampaigns($dispositionGroup, 'update');
        
        if ($response) return $response;
        
        $dispositionGroup->update($validatedData);

        $dispositions = collect($validatedData['dispositions']);

        $dispositionIds = $dispositionGroup->dispositions()->pluck('id')->toArray();

        $currentDispositions = $dispositions->whereIn('id', $dispositionIds);

        $dispositionGroup->dispositions()->whereNotIn('id', $currentDispositions->pluck('id')->toArray())->delete();

        $currentDispositions->each(function ($item) use ($dispositionGroup) {
            $dispositionGroup->dispositions()->updateOrCreate(
                ['id' => $item['id']],
                ['name' => $item['name'], 'category' => $item['category'],
                 'tenant_id' => $dispositionGroup->tenant_id, 'reasons' => $item['reasons']
                ]
            );
        });
        
        $dispositions->transform(function ($item) use ($validatedData) {
            $item['tenant_id'] = $validatedData['tenant_id'];
            return $item;
        });

        $newDispositions = $dispositions->whereNotIn('id', $dispositionIds);

        $dispositionGroup->dispositions()->createMany($newDispositions);
        
        $dispositionGroup->load('dispositions');
        
        return ['data' => $dispositionGroup];
    }

    public function destroy(Request $request, $id)
    {
        $dispositionGroup = DispositionGroup::with('campaigns')->findOrFail($id);

        $this->authorize('updateDestroy', $dispositionGroup);

        $response = $this->checkCampaigns($dispositionGroup, 'delete');

        if ($response) return $response;
        
        $dispositionGroup->delete();
        
        return response()->json([], 204);
    }
    
    private function checkCampaigns($dispositionGroup, $action)
    {
        $message = [
            'update' => __('disposition_group_cannot_update'),
            'delete' => __('disposition_group_cannot_delete')
        ];
        $activeCampaigns = $dispositionGroup->campaigns()->active();
        if ($activeCampaigns->count() > 0) {
            return response()->json(['message' => $message[$action] . ': ' . $activeCampaigns->pluck('name')->implode(', ') ], 422);
        }
    }
}
