<?php

namespace App\Http\Controllers\Api;

use App\Exports\DncNumbersExport;
use App\Imports\DncNumbersImport;
use App\Models\DncNumber;
use App\Http\Requests\DncRequest;
use App\Http\Controllers\Controller;
use App\Models\User;
use App\Utils\TenantUtils;
use Illuminate\Http\Request;
use Maatwebsite\Excel\Facades\Excel;
use Maatwebsite\Excel\Validators\ValidationException;

class DncController extends Controller
{
    public function index(Request $request)
    {
        $user = auth()->user();

        // Determine tenant ID
        $tenantId = $user->tenant_id;
        if ($user->hasRole('super-admin') && $request->filled('tenant_id')) {
            $tenantId = $request->input('tenant_id');
        }

        // Pagination & Sorting Parameters
        $perPage = (int) $request->input('per_page', 10);
        $search = $request->input('search', '');
        $sortBy = $request->input('sort_by', 'id');
        $sortOrder = $request->input('sort_order', 'desc');

        // Whitelist allowed sortable columns to prevent SQL injection
        $allowedSortColumns = ['id', 'phone_number', 'reason', 'created_at'];
        if (!in_array($sortBy, $allowedSortColumns)) {
            $sortBy = 'id';
        }

        // Base query
        $query = DncNumber::where('tenant_id', $tenantId)->with('addedBy');

        // Apply search filter
        if (!empty($search)) {
            $query->where(function ($q) use ($search) {
                $q->where('phone_number', 'like', "%$search%")
                    ->orWhere('reason', 'like', "%$search%")
                    ->orWhereHas('addedBy', function ($subQuery) use ($search) {
                        $subQuery->where('name', 'like', "%$search%");
                    });
            });
        }

        // Apply sorting
        $query->orderBy($sortBy, $sortOrder);

        // Return paginated results
        return $query->paginate($perPage);
    }

    public function store(DncRequest $request)
    {
        $validated = $request->validated();
        $this->authorize('create', [
            DncNumber::class,
            $validated['tenant_id'],
        ]);
        $dncNumber = DncNumber::create($validated);
        $dncNumber->addedBy()->associate(auth()->user());
        $dncNumber->save();
        return $dncNumber;
    }

    public function destroy(DncNumber $dnc)
    {
        $this->authorize('viewUpdateDestroy', $dnc);

        $dnc->delete();
        return response()->noContent();
    }

    public function import (Request $request) {
        $request->validate([
            'file' => 'required|mimes:csv',
        ]);

        $file = $request->file('file');
        $user = auth()->user();
        try {
            Excel::import(new DncNumbersImport(TenantUtils::getTenantId($request), $user), $file);
        } catch (ValidationException $e) {
            return response()->json(['failures' => $e->failures()], 422);
        }

        return response()->json([]);

    }

    // create a new method for exporting the dnc numbers, get the tenant_id and the format type from the request
    public function export(Request $request)
    {
        $tenantId = TenantUtils::getTenantId($request);
        $format = $request->input('format', 'csv');

        if ($format === 'csv') {
            $format = \Maatwebsite\Excel\Excel::CSV;
        } elseif ($format === 'xlsx') {
            $format = \Maatwebsite\Excel\Excel::XLSX;
        } elseif ($format === 'xls') {
            $format = \Maatwebsite\Excel\Excel::XLS;
        }

        // Generate the filename based on the current date, tenant id and time and the selected format
        $fileName = 'dnc-numbers-' . date('Y-m-d') . '-' . $tenantId . '-' . time() . '.' . $format;

        return Excel::download(new DncNumbersExport(
            TenantUtils::getTenantId($request),
            $format
        ), $fileName, $format);
    }
}
