<?php

namespace App\Http\Controllers\Api;

use App\Http\Controllers\Controller;
use App\Http\Requests\Queues\StoreQueue;
use App\Models\Campaign;
use App\Models\User;
use App\Utils\{TenantUtils, VitalPBXApiUtils};
use Illuminate\Http\Request;

class QueueController extends Controller
{
    public function index(Request $request)
    {

	    /** @var User $user */
        $user = auth()->user();
	    $user->load('roles', 'systemPermission');

        $tenantId = TenantUtils::getTenantId($request, $user);

        $queues = VitalPBXApiUtils::request('queues/outbound', $tenantId);

        if ($user->hasRole('super-admin') || is_null($user->systemPermission) ||
            ($user->hasPermission('system permissions') && $request->input('all') === 'true')) {
            return $queues;
        }

        //filter queues by system permissions
        $collection = collect($queues['data'] ?? []);

        $systemPermissionQueues = collect($user->systemPermission->queues)->map(function ($values) {
            return $values['name'];
        });

        $collection = $collection->whereIn('name', $systemPermissionQueues);

        return ['data' => $collection->values()->all()];
    }

    public function store(StoreQueue $request)
    {
        $validatedData = $request->validated();

        $tenant = TenantUtils::getTenantId($request);

        $response = VitalPBXApiUtils::request('queues', $tenant, 'POST', $this->params($validatedData));
        return $this->createUpdateResponse($response, $tenant);
    }

    public function update(StoreQueue $request, $id)
    {
        $validatedData = $request->validated();

        $tenant = TenantUtils::getTenantId($request);

        $response = VitalPBXApiUtils::request("queues/{$id}", $tenant, 'PUT', $this->params($validatedData));

        return $this->createUpdateResponse($response, $tenant);
    }

    public function destroy(Request $request, $id)
    {
        //validates if queue is associated to active campaign before delete
        if (Campaign::active()->where('queue_id', $id)->count() > 0) {
            return response()->json([
				'message' => __('queues_cannot_delete_with_active_campaign')
            ], 422);
        }

        $tenant = TenantUtils::getTenantId($request);

        $response = VitalPBXApiUtils::request("queues/{$id}", $tenant, 'DELETE');

        if (is_array($response) && empty($response['data'])) $this->applyChangesRequest($tenant);

        return $response;
    }

    private function createUpdateResponse($response, $tenant)
    {
        if (!is_array($response) && $response->status() !== 200) {
            return $response;
        }
        if (isset($response['data'])) $this->applyChangesRequest($tenant);

        return $response;
    }

    private function params($validatedData)
    {
        return [
            'extension' => $validatedData['extension'],
            'queue_vip_list_id' => null,
            'ivr_id' => null,
            'queue_callback_id' => null,
            'cron_profile_id' => null,
            'description' => $validatedData['description'],
            'prefix' => null,
            'direction' => 'outbound',
            'dial_mode' => $validatedData['dial_mode'],
            'strategy' => 'rrordered',
            'music_group_id' => 1,
            'weight' => 0,
            'autofill' => true,
            'maxlen' => 0,
            'announcement_id' => null,
            'periodic_announcement_id' => 1,
            'join_announcement_id' => null,
            'record' => false,
            'servicelevel' => 30,
            'wrapuptime' => 0,
            'announce_frequency' => null,
            'min_announce_frequency' => null,
            'periodic_announce_frequency' => 15,
            'announce_round_seconds' => 0,
            'announce_to_first_user' => false,
            'announce_position' => 'no',
            'announce_position_limit' => null,
            'relative_periodic_announce' => false,
            'announce_holdtime' => false,
            'penaltymemberslimit' => null,
            'autopause' => 'no',
            'ringinuse' => false,
            'memberdelay' => null,
            'timeoutrestart' => false,
            'joinempty' => 'yes',
            'timeoutpriority' => 'app',
            'timeout' => 15,
            'queue_timeout' => 0,
            'leavewhenempty' => 'no',
            'retry' => 5,
            'answered_elsewhere' => false,
            'alertinfo' => null,
            'answerchannel' => true,
            'members' => $this->getMembers($validatedData['members']),
            'final_destination' => [
                'category_id' => 24,
                'destination_id' => 1
            ],
            'after_hangup_destination' => [
                'category_id' => null,
                'destination_id' => null
            ]
        ];
    }

    private function getMembers($members)
    {
        $rows = collect();

        foreach($members as $item) {
            $rows->push([
                'extension_id' => $item,
                'penalty' => 0,
                'diversions' => true,
                'type' => 'dynamic'
            ]);
        }

        return $rows->all();
    }

    private function applyChangesRequest($tenant)
    {
        return VitalPBXApiUtils::request("tenants/{$tenant}/apply_changes", $tenant, 'UPDATE', null, null, 120);
    }
}
