<?php

namespace App\Http\Controllers\Api;

use App\Exports\UsersExport;
use App\Http\Controllers\Controller;
use App\Http\Requests\Users\{StoreUser, UpdateUser};
use App\Http\Resources\User as UserResource;
use App\Imports\UsersImport;
use App\Models\User;
use App\Utils\{LicenseUtils, TenantUtils, VitalPBXApiUtils};
use Illuminate\Http\Request;
use Maatwebsite\Excel\Facades\Excel;
use Maatwebsite\Excel\Validators\ValidationException;

class UserController extends Controller
{
    public function index(Request $request)
    {
        $this->authorize('viewAny', User::class);

        $user = auth()->user()->load('systemPermission');

        return UserResource::collection(User::with('roles')
            ->when($request->input('agents'), function ($query) {
                return $query->agents();
            })
            ->when(($request->input('agents') && $user->systemPermission), function ($query) use ($user) {
                return $query->extensions($user->systemPermission->agents);
            })
            ->byTenant(TenantUtils::getTenantId($request))
            ->orderBy('updated_at', 'desc')
            ->get());
    }

    public function store(StoreUser $request)
    {
        $validatedData = $request->validated();

        $this->authorize('users', User::class);

        $createAgent = $this->isAgent($validatedData['roles']);

        //verify if more user agents can be created
        if ($createAgent && ! LicenseUtils::checkLicenseRestriction()) {
            return response()->json(['message' => __('agent_limit')], 422);
        }

        $validatedData['password'] = bcrypt($validatedData['password']);

        $user = User::create($validatedData);

        $user->roles()->attach($this->getRoleId($validatedData['roles']));

        if ($createAgent) {
            //add default user settings for new agent
            $user->settings()->createMany([
                ['type' => 'audio', 'key' => 'echo_cancellation', 'value' => 1],
                ['type' => 'audio', 'key' => 'noise_suppression', 'value' => 1],
                ['type' => 'calls_settings', 'key' => 'auto_answer', 'value' => 0],
                ['type' => 'calls_settings', 'key' => 'ringtone', 'value' => 'telephone_ring_variant_1.mp3']
            ]);
        }

        return new UserResource($user);
    }

    public function show(Request $request, $id)
    {
        $user = User::with('roles')->findOrFail($id);

        $this->authorize('viewUpdateDestroy', $user);

        return new UserResource($user);
    }

    public function update(UpdateUser $request, User $user)
    {
        $validatedData = $request->validated();

        $this->authorize('viewUpdateDestroy', $user);

        $this->updateExtensionId($validatedData, $user);

        if (empty($validatedData['password'])) {
            unset($validatedData['password']);
        } else {
            $validatedData['password'] = bcrypt($validatedData['password']);
        }

        //system permission set to null if user tenant id is changed
        if (! $user->isSuperAdmin()
            && $user->tenant_id != $validatedData['tenant_id']
            && $validatedData['system_permission_id'] == $user->system_permission_id) {
            $validatedData['system_permission_id'] = null;
        }

        $user->update($validatedData);

        $user->roles()->sync($this->getRoleId($validatedData['roles']));

        return new UserResource($user);
    }

    public function destroy(Request $request, User $user)
    {
        $this->authorize('viewUpdateDestroy', $user);

        if ($user->id === auth()->user()->id) return response()->json(['message' => __('delete_current_user')], 422);

        // if ($user->isSuperAdmin()) return response()->json(['message' => __('delete_admin_user')], 422);

        $user->delete();

        return response()->json([], 204);
    }

    public function import(Request $request)
    {
        $validatedData = $request->validate([
            'file' => 'required|file',
            'tenant_id' => 'required'
        ]);

        $this->authorize('users', User::class);

        try {
            Excel::import(new UsersImport(TenantUtils::getTenantId($request)), $validatedData['file']);
        } catch (ValidationException $e) {
            return response()->json(['failures' => $e->failures()], 422);
        }

        return response()->json([]);
    }

    public function export(Request $request)
    {
        $this->authorize('users', User::class);

        return Excel::download(new UsersExport(TenantUtils::getTenantId($request)), 'import-users-template.csv');
    }

    /**
     * return each role id from array to associate to user
     */
    private function getRoleId($roles)
    {
        return collect($roles)->map(function ($item) {
            return $item['id'];
        });
    }

    private function isAgent($roles)
    {
        return collect($roles)->contains('name', 'agent');
    }

    //if extension_id is changed update the related queues
    private function updateExtensionId($validatedData, $user)
    {
        if ($this->isAgent($validatedData['roles']) && $validatedData['extension_id'] !== $user->extension_id) {
            $response = VitalPBXApiUtils::request('queues/outbound', $validatedData['tenant_id']);

            if (is_array($response)) {
                $queues = collect($response['data']);
                $queues->each(function ($item) use ($user, $validatedData) {

                    [$members, $update] = $this->getMembers($item['members'], $user, $validatedData);

                    //send request only if member is in queue
                    if ($update) {
                        $data = $this->queueParams($item, $members);

                        VitalPBXApiUtils::request("queues/{$item['queue_id']}", $validatedData['tenant_id'], method: 'PUT', params: $data);
                    }
                });

                VitalPBXApiUtils::request("tenants/{$validatedData['tenant_id']}/apply_changes", $validatedData['tenant_id'], method: 'UPDATE');
            } else {
                return $response;
            }
        }
    }

    private function queueParams($queue, $members)
    {
        return [
            'extension' => $queue['extension'],
            'queue_vip_list_id' => null,
            'ivr_id' => null,
            'queue_callback_id' => null,
            'cron_profile_id' => null,
            'description' => $queue['description'],
            'prefix' => null,
            'direction' => 'outbound',
            'dial_mode' => $queue['dial_mode'],
            'strategy' => 'rrordered',
            'music_group_id' => 1,
            'weight' => 0,
            'autofill' => true,
            'maxlen' => 0,
            'announcement_id' => null,
            'periodic_announcement_id' => 1,
            'join_announcement_id' => null,
            'record' => false,
            'servicelevel' => 30,
            'wrapuptime' => 0,
            'announce_frequency' => null,
            'min_announce_frequency' => null,
            'periodic_announce_frequency' => 15,
            'announce_round_seconds' => 0,
            'announce_to_first_user' => false,
            'announce_position' => 'no',
            'announce_position_limit' => null,
            'relative_periodic_announce' => false,
            'announce_holdtime' => false,
            'penaltymemberslimit' => null,
            'autopause' => 'no',
            'ringinuse' => false,
            'memberdelay' => null,
            'timeoutrestart' => false,
            'joinempty' => 'yes',
            'timeoutpriority' => 'app',
            'timeout' => 15,
            'queue_timeout' => 0,
            'leavewhenempty' => 'no',
            'retry' => 5,
            'answered_elsewhere' => false,
            'alertinfo' => null,
            'answerchannel' => true,
            'members' => $members,
            'final_destination' => [
                'category_id' => 24,
                'destination_id' => 1
            ],
            'after_hangup_destination' => [
                'category_id' => null,
                'destination_id' => null
            ]
        ];
    }

    //replace the extension id value to update
    private function getMembers($members, $user, $validatedData)
    {
        $rows = collect();
        $update = false;

        foreach($members as $item) {
            if ($item['extension_id'] === $user->extension_id) {
                $extensionId = $validatedData['extension_id'];
                $update = true;
            } else {
                $extensionId = $item['extension_id'];
            }

            $rows->push([
                'extension_id' => $extensionId,
                'penalty' => 0,
                'diversions' => true,
                'type' => 'dynamic'
            ]);
        }

        return [$rows->all(), $update];
    }
}
