<?php

namespace App\Jobs;

use App\Models\{BroadcastCampaign, BroadcastCampaignContact, TimezoneSetting};
use App\Utils\VitalPBXApiUtils;
use Illuminate\Bus\Queueable;
use Illuminate\Contracts\Queue\ShouldBeUnique;
use Illuminate\Contracts\Queue\ShouldQueue;
use Illuminate\Foundation\Bus\Dispatchable;
use Illuminate\Queue\InteractsWithQueue;
use Illuminate\Queue\SerializesModels;
use Illuminate\Support\Facades\Log;

class CallBroadcastCampaignContacts implements ShouldQueue, ShouldBeUnique
{
    use Dispatchable, InteractsWithQueue, Queueable, SerializesModels;

    /**
     * The number of times the job may be attempted.
     *
     * @var int
     */
    public int $tries = 5;

    /**
     * The maximum number of unhandled exceptions to allow before failing.
     *
     * @var int
     */
    public int $maxExceptions = 3;

    /**
     * The number of seconds the job can run before timing out.
     *
     * @var int
     */
    public int $timeout = 3600;

    /**
     * The number of seconds to wait before retrying the job.
     *
     */
    public array $backoff = [5, 10, 15, 20, 25];

    /**
     * The podcast instance.
     *
     * @var BroadcastCampaign
     */
    public BroadcastCampaign $campaign;

    public $timezone = null;

    /**
     * Create a new job instance.
     * @param BroadcastCampaign $campaign
     * @return void
     */
    public function __construct(BroadcastCampaign $campaign)
    {
        $this->campaign = $campaign;
        $this->timezone = TimezoneSetting::byTenant($campaign->tenant_id)->first();
        if ($this->timezone) {
            $this->timezone = $this->timezone->timezone;
        }
    }

    /**
     * The unique ID of the job.
     *
     */
    public function uniqueId()
    {
        return $this->campaign->id;
    }

    /**
     * Execute the job.
     *
     * @return void
     */
    public function handle(): void
    {
        // Double-check the campaign status before dispatching contact jobs
        $campaign = BroadcastCampaign::find($this->campaign->id);

        if (!$this->isActive($campaign)) {
            return; // Stop processing if the campaign is no longer active
        }

        // Log::info('Dispatching contact job for campaign ' . $campaign->name . ' with status ' . $campaign->status);

        $contacts = BroadcastCampaignContact::with('contact')
            ->pending($this->campaign->id)
            ->get();
        $chunkedContacts = $contacts->chunk(100);
        $callsPerMinute = $this->campaign->calls_per_minute;
        $interval = 60 / $callsPerMinute * 1000; // Target interval between calls in milliseconds
        $startTime = microtime(true);

        foreach ($chunkedContacts as $chunk) {
            foreach ($chunk as $index => $broadcastCampaignContact) {
                $this->campaign->refresh();
                // Double-check the campaign status before dispatching CallContact
                if (!$this->isActive($this->campaign)) {
                    return; // Stop processing if the campaign is no longer active
                }
                // Dispatch the CallContact job
                CallContact::dispatch($broadcastCampaignContact, $this->campaign);
                // Calculate the wait time before the next call to fit the target interval
                $elapsedTime = microtime(true) - $startTime;
                $waitTime = max(0, ($interval - ($elapsedTime * 1000))); // Time Remaining for the Next Interval
                usleep($waitTime * 1000); // Convert from milliseconds to microseconds
                $startTime = microtime(true); // Update the start time for the next iteration
            }
        }

        $broadcastCampaignContacts = BroadcastCampaignContact::with('contact')
            ->pending($this->campaign->id)
            ->get();
        if ($broadcastCampaignContacts->count() < 1) {
            $this->campaign->update(['status' => 'finished']);
        }
    }

    /**
     * @param BroadcastCampaign $campaign
     * @return bool
     */
    public function isActive(BroadcastCampaign $campaign): bool
    {
        $now = $this->timezone ? now($this->timezone) : now();

        return $campaign->status === 'active'
            && $campaign->start_date <= $now
            && $campaign->end_date >= $now
            && $now->between($campaign->start_shift, $campaign->end_shift);
    }

    public function callContact (BroadcastCampaignContact $broadcastCampaignContact, $index) {
        try {
            // define how long to wait for call to be answered (in secs.)
            $timeout = 30;
            $params = [
                "number" => $broadcastCampaignContact->contact->phone_1,
                "cos_id" => $this->broadcastCampaign->class_of_service_id,
                "destination_category_id" => $this->broadcastCampaign->destination_category_id,
                "destination_id" => $this->broadcastCampaign->destination_id,
                "cid_name" => $this->broadcastCampaign->cid_name,
                "cid_number" => $this->broadcastCampaign->cid_number,
                "timeout" => $timeout,
            ];

            $httpRequestTimeout = $timeout + 10;

            $response = VitalPBXApiUtils::request(
                'core/dialer_call',
                $this->broadcastCampaign->tenant_id,
                method: 'POST',
                params: $params,
                timeout: $httpRequestTimeout
            );
            $timezone = TimezoneSetting::byTenant($this->broadcastCampaign->tenant_id)->first();
            $now = $timezone ? now($timezone->timezone) : now();
            // Log::info( 'CallBroadcastCampaignContacts response: ' . json_encode( $response, JSON_THROW_ON_ERROR ) );
            if (!is_array($response) && $response->status() !== 200) {
                $broadcastCampaignContact->update([
                    'status' => 'failed',
                    'called_at' => $now,
                ]);
            } else {
                $broadcastCampaignContact->update([
                    'status' => 'finished',
                    'called_at' => $now,
                ]);
            }
        } catch (\Exception $e) {
            Log::error($e->getMessage());
        }
    }

}
