<?php

namespace App\Jobs;

use App\Models\BroadcastCampaign;
use App\Models\BroadcastCampaignContact;
use App\Models\TimezoneSetting;
use App\Utils\VitalPBXApiUtils;
use Illuminate\Bus\Queueable;
use Illuminate\Contracts\Queue\ShouldBeUnique;
use Illuminate\Contracts\Queue\ShouldQueue;
use Illuminate\Foundation\Bus\Dispatchable;
use Illuminate\Queue\InteractsWithQueue;
use Illuminate\Queue\Middleware\WithoutOverlapping;
use Illuminate\Queue\SerializesModels;

class CallContact implements ShouldQueue, ShouldBeUnique
{
    use Dispatchable, InteractsWithQueue, Queueable, SerializesModels;

    public BroadcastCampaignContact $contact;
    public BroadcastCampaign $campaign;

    /**
     * Create a new job instance.
     *
     * @param BroadcastCampaignContact $contact
     * @param \App\Models\BroadcastCampaign $campaign
     */
    public function __construct(BroadcastCampaignContact $contact, BroadcastCampaign $campaign)
    {
        $this->contact = $contact;
        $this->campaign = $campaign;
    }

    /**
     * The unique ID of the job.
     *
     */
    public function uniqueId()
    {
        return $this->contact->id;
    }

    // WithoutOverlapping prevents jobs with the same key from running at the same time.
    // dontRelease() will delete any new job attempts if a job with the same key is currently running.
    public function middleware(): array
    {
        return [(new WithoutOverlapping($this->contact->id))->dontRelease()];
    }

    /**
     * Execute the job.
     *
     * @return void
     */
    public function handle(): void
    {
        // Define how long to wait for the call to be answered (in seconds)
        $timeout = 30;
        $params = [
            "number" => $this->contact->contact->phone_1,
            "cos_id" => $this->campaign->class_of_service_id,
            "destination_category_id" => $this->campaign->destination_category_id,
            "destination_id" => $this->campaign->destination_id,
            "cid_name" => $this->campaign->cid_name,
            "cid_number" => $this->campaign->cid_number,
            "timeout" => $timeout,
        ];

        $httpRequestTimeout = $timeout + 10;

        $response = VitalPBXApiUtils::request(
            'core/dialer_call',
            $this->campaign->tenant_id,
            method: 'POST',
            params: $params,
            timeout: $httpRequestTimeout
        );

        $timezone = TimezoneSetting::byTenant($this->campaign->tenant_id)->first();
        $now = $timezone ? now($timezone->timezone) : now();

        if (!is_array($response) && $response->status() !== 200) {
            $this->contact->update([
                'status' => 'failed',
                'called_at' => $now,
            ]);
        } else {
            $this->contact->update([
                'status' => 'finished',
                'called_at' => $now,
            ]);
        }
    }
}
