<?php

namespace App\Jobs;

use App\Models\BroadcastCampaign;
use App\Utils\VitalPBXApiUtils;
use Illuminate\Bus\Batchable;
use Illuminate\Bus\Queueable;
use Illuminate\Contracts\Queue\ShouldBeUnique;
use Illuminate\Contracts\Queue\ShouldQueue;
use Illuminate\Foundation\Bus\Dispatchable;
use Illuminate\Http\Client\ConnectionException;
use Illuminate\Queue\InteractsWithQueue;
use Illuminate\Queue\SerializesModels;
use Illuminate\Support\Facades\Http;
use Illuminate\Support\Facades\Log;

class ProcessBroadcastCampaignContactJob implements ShouldQueue, ShouldBeUnique
{

    use Batchable, Dispatchable, InteractsWithQueue, Queueable, SerializesModels;

    /**
     * The number of seconds after which the job's unique lock will be released.
     *
     * @var int
     */
    public int $uniqueFor = 3600; // Prevent duplicates for 1 hour

    /**
     * Delete the job if its models no longer exist.
     *
     * @var bool
     */
    public bool $deleteWhenMissingModels = true;

    public function __construct(protected $broadcastContact, protected int $broadcastCampaignId, protected mixed $timezone = null) {}

    public function uniqueId(): string
    {
        return "{$this->broadcastCampaignId}-{$this->broadcastContact->id}";
    }

    public function handle()
    {
        if ($this->batch() && $this->batch()->cancelled()) {
            // Determine if the batch has been cancelled...
            return;
        }
        // Check if the campaign is still active
        $broadcastCampaign = BroadcastCampaign::find($this->broadcastCampaignId);

        if (!$broadcastCampaign || $broadcastCampaign->status !== 'active') {
            if ($this->batch()) {
                $this->batch()->cancel();
                // Log::info('Campaign has been cancelled');
            }
            $this->broadcastContact->update([
                'status' => 'pending',
                'called_at' => null
            ]);
            // Log::info("Broadcast campaign {$this->broadcastCampaignId} is no longer active. Stopping job for contact {$this->broadcastContact->contact_id}.");
            return;
        }

        if (!$this->broadcastContact || $this->broadcastContact->status !== 'queued') {
            return;
        }

        $contact = $this->broadcastContact->contact;

        // Define how long to wait for the call to be answered (in seconds)
        $timeout = 30;
        $params = [
            "number" => $contact->phone_1,
            "cos_id" => $broadcastCampaign->class_of_service_id,
            "destination_category_id" => $broadcastCampaign->destination_category_id,
            "destination_id" => $broadcastCampaign->destination_id,
            "cid_name" => $broadcastCampaign->cid_name,
            "cid_number" => $broadcastCampaign->cid_number,
            "timeout" => $timeout,
        ];

        $httpRequestTimeout = $timeout + 2;

        try {
            $response = VitalPBXApiUtils::request(
                'core/dialer_call',
                $broadcastCampaign->tenant_id,
                method: 'POST',
                params: $params,
                timeout: 1
            );

            $now = $this->timezone ? now($this->timezone->timezone) : now();

            if (!is_array($response) && $response->status() !== 200) {
                $this->broadcastContact->update([
                    'status' => 'failed',
                    'called_at' => $now,
                ]);
            } else {
                $this->broadcastContact->update([
                    'status' => 'finished',
                    'called_at' => $now,
                ]);
            }
        } catch (ConnectionException $e) {
            $now = $this->timezone ? now($this->timezone->timezone) : now();
            $this->broadcastContact->update([
                'status' => 'finished',
                'called_at' => $now,
            ]);
        } catch (\Exception $e) {
            // Log::error("Failed to process contact {$this->broadcastContact->id}: {$e->getMessage()}");
            $now = $this->timezone ? now($this->timezone->timezone) : now();
            $this->broadcastContact->update([
                'status' => 'failed',
                'called_at' => $now,
            ]);
        }
    }
}